from ase.calculators.calculator import Calculator
from ase.calculators.singlepoint import SinglePointCalculator
from ase.io import write,Trajectory
from sklearn.cluster import KMeans
import numpy as np
import time
import copy

class FALCON(Calculator):
    """
    On-the-Fly Machine Learning Calculator

    Parameters
    -----------
    model : falcon-md.models.model_base_class.ModelBaseClass or list of Models
        ML model to use for the OTF training or list of models to load.
    
    calc : ASE calculator
        ASE calculator object to perform exact calculations.

    training_data : list of ASE atoms objects
        List of structures for initial training of the ML model.

    accuracy_e : float
        Accuracy threshold (Epsilon) in eV. Default: 0.1

    accuracy_f : float, optional
        Optional accuracy threshold for forces in eV/Å. Default: inf

    accuracy_update : float, optional
        Optional value by which the accuracy threshold is updated in every step in eV. Default: 0.0

    modelsize : int
        Average model size, defines number of structures before a new ML model is added. Default: 500

    max_clusters : int
        Maximum of clusters/ML models. Default: 1

    start_clusters : int
        Number of clusters/ML models used at the start of the simulation. Default: 1
    
    train_start : int
        Number of initial predictions, where the ML model is trained regardless of the uncertainty. Default: 10

    train_every : int
        Periodic retraining interval, regardless of uncertainty. Default: 1e9

    train_log : bool
        If True, retraining occurs at exponentially increasing intervals (2**x steps). Default: True.

    i_batch : int
        Defines number of structures added to training data of one model before batch training is performed. Default: 1

    write_training_data : bool
        Save training structures to file during simulation. Default: True.

    td_filename : str
        Filename for training data output. Default: "OTF_Training_Data.traj".


    """

    
    implemented_properties = ['energy', 'forces']
    default_parameters = {}



    def __init__(self,
                 model,
                 calc,
                 training_data, 
                 accuracy_e=0.1, 
                 accuracy_f=float('inf'),
                 accuracy_update=0.0, 
                 modelsize=500, 
                 max_clusters=1,
                 start_clusters=1, 
                 train_start=10, 
                 train_every=1e9,
                 train_log=True, 
                 i_batch=1,
                 write_training_data=True,  
                 td_filename = 'OTF_Training_Data.traj', 
                 **kwargs):

        print("If you are using FALCON in a publication, please cite: \n Felis, N., Dononelli, W. (2025). FALCON: fast active learning for machine learning potentials in atomistic and ab initio molecular dynamics simulations. npj Comput. Mater., https://doi.org/10.1038/s41524-025-01897-8")

        Calculator.__init__(self, **kwargs)

        self.calc = calc
        self.steps = 0               # MD step count 
        self.training_steps = 0      # Count of performed retrainings.
        self.initialize = True
        self.accuracy_e = accuracy_e
        self.accuracy_f = accuracy_f
        self.accuracy_update = accuracy_update
        self.training_data = training_data
        self.train_start = train_start
        self.train_every = train_every
        self.write_training_data = training_data
        self.training_data_filename = td_filename 
        self.modelsize = modelsize
        self.start_clusters = start_clusters
        self.log_test = (lambda x: (np.log2(x) % 1) == 0) if train_log else (lambda x: False)
        self.i_batch = i_batch

        # Setup of the ML models
        if not type(model)==list:
            self.models = []
            self.active_models = []
            self.descriptor = model.descriptor
            self.last_clustering = 0
            self.n_clusters = 0
            self.max_clusters = max_clusters
            self.reloaded = False

        # Creation of number of  ML models that is defined with max_clusters.
            for i in range(max_clusters):
                model_copy = copy.deepcopy(model)
                setattr(self, f"model{i}", model_copy)
                self.models.append(getattr(self, f"model{i}"))

        else:
            self.models = model
            self.active_models = model
            self.descriptor = model[0].descriptor
            self.last_clustering = 0
            self.n_clusters = len(model)
            self.max_clusters = len(model)
            self.reloaded = True





    def calculate(self, atoms=None, properties=['energy', 'forces'], system_changes=['positions']):
        Calculator.calculate(self, atoms, properties, system_changes)

        print(f'\n============================ Step: {self.steps+1} ============================\n')
    
        print(f'Total structures in training data: {len(self.training_data)}')


        # Clustering of training data into subsets of data (triggered if modelsize is reached or self.initialize == True)

        if ((len(self.training_data) % self.modelsize == 0 and self.n_clusters <= self.max_clusters) or self.initialize) and len(self.training_data) != self.last_clustering:
            self.last_clustering = len(self.training_data)

            if self.initialize:
                self.n_clusters = self.start_clusters - 1

            if self.n_clusters < self.max_clusters:
                start=time.time()
                self.n_clusters +=1
                while len(self.active_models)<self.n_clusters:
                    new_model = self.models.pop(0)
                    self.active_models.append(new_model)
                self.batch_counts = [0] * len(self.active_models)
                print(f'{len(self.active_models)} active Models.')

                # Cluster all structures in training data using KMeans
                X = self.get_features(self.training_data)
                self.kmeans = KMeans(n_clusters=self.n_clusters, init="k-means++", n_init=10, random_state=np.random.randint(0, 10e6)).fit(X)
                self.labels = self.kmeans.labels_
                print(self.labels)
                self.clusters = {i: [] for i in range(self.n_clusters)}

                # Assign structures to their ML model.
                for structure, label in zip(self.training_data, self.labels):
                    self.clusters[label].append(structure)

                for cluster_id, structures in self.clusters.items():
                    print(f"Cluster {cluster_id} contains {len(structures)} objects.")
                end = time.time()
                t = end-start
                print(f"Clustering of {len(self.training_data)} took {t} s.")

                # Training of all active ML models.
                start = time.time()
                for i, model in enumerate(self.active_models):
                    start_model = time.time()
                    print(type(model))
                    model.train(self.clusters[i])
                    end_model = time.time()
                    t_model = end_model - start_model
                    print(f"Model {i} trained with {len(self.clusters[i])} structures ({t_model} s).")
                    print()
                end = time.time()
                t = end-start
                if self.initialize:
                    if self.write_training_data:
                        with Trajectory(self.training_data_filename, mode='w') as traj:
                            for structure in self.training_data:
                                traj.write(structure)
                    print(f'Model trained with initial training data. ({t} s)')
                    self.initialize = False

                else:
                    print(f'Step {self.steps}: Clustering completed. {len(self.active_models)} models trained. ({t} s)')
                print()
                

        # Prediction of energy, forces and uncertainties
        E = None
        F = None
        Fmax = None
        Fstdmax = float('inf')
        best_model = None
        Estd = float('inf')
        for i, model in enumerate(self.active_models):
            print(f'Model {i}...')
            start = time.time()
            E_model = model.predict_energy(atoms)
            end = time.time()
            t = end-start
            print(f'Predict energy ({t} s)')

            start = time.time()
            Estd_model = model.predict_uncertainty(atoms)
            end = time.time()
            t = end-start
            print(f'Predict energy_uncertainty ({t} s)')
            start = time.time()
            F_model = model.predict_forces(atoms)
            Fmax_model = np.max(np.linalg.norm(F_model, axis=1))
            end = time.time()
            t = end-start
            print(f'Predict forces ({t} s)')

            start = time.time()
            Fstd_model = model.predict_uncertainty_forces(atoms)
            Fstdmax_model = np.max(np.linalg.norm(Fstd_model, axis=1))
            end = time.time()
            t = end-start
            print(f'Predict forces_uncertainty ({t} s)')


            print(f'E = {E_model} (Model {i})')
            print(f'Estd = {Estd_model} (Model {i})')
            print(f'F = {Fmax_model} (Model {i})')
            print(f'Fstd = {Fstdmax_model} (Model {i})')
            print()

             # Select model with lowest uncertainty
            if Estd_model<Estd:
                Estd = Estd_model
                E = E_model
                F = F_model
                Fmax = Fmax_model
                Fstdmax = Fstdmax_model
                best_model = i

        print(f'Model {best_model} has the lowest uncertainty.')
        print(f'E = {E}')
        print(f'Estd = {Estd}')
        print(f'Fmax = {Fmax}')
        print(f'Fstd = {Fstdmax}')
        print()
        
        # Determine if an exact calculation is necessary and perform it
        if (self.steps<self.train_start or self.steps % self.train_every == 0 or self.log_test(self.steps) or Estd>self.accuracy_e or Fstdmax>self.accuracy_f) and self.reloaded==False:
           print('Exact calculation...')           
           b=atoms.copy()
           b.calc=copy.deepcopy(self.calc)
           start = time.time()
           E=b.get_potential_energy()
           F=b.get_forces()
           end = time.time()
           t = end-start
           self.steps += 1
           print(f'Exact Calculation ({t} s)')
           print(f'Real energy = {E}')
           print(f'Real Fmax = {np.max(np.linalg.norm(F, axis=1))}')

           self.training_data.append(b)

           # Append new training structure to OTF training data file.
           if self.write_training_data:
                with Trajectory(self.training_data_filename, mode='a') as traj:
                    traj.write(b)

           

           # Predict nearest cluster of new training structure
           Y = self.descriptor.create_features(b)
           pred_cluster = self.kmeans.predict(Y)
           print(f'New structure is added to Model {pred_cluster[0]}.')
           print()
           self.clusters[pred_cluster[0]].append(b)
           self.batch_counts[pred_cluster[0]] += 1

           # Only retrain the updated ML model
           if self.batch_counts[pred_cluster[0]] >= self.i_batch and self.steps>=self.train_start:
               start = time.time()
               self.active_models[pred_cluster[0]].train(self.clusters[pred_cluster[0]])
               end = time.time()
               t = end-start
               self.training_steps += 1
               self.batch_counts[pred_cluster[0]] = 0
               print(f'Step {self.steps}: Model {pred_cluster[0]} trained with {len(self.clusters[pred_cluster[0]])} Structures ({t} s). ({self.training_steps} Training steps.) ')
               print()

           else:
               print(f'Step {self.steps}: New structure added to training data of Model {pred_cluster[0]}. No training is performed since number of new structures for batch training is not reached yet.')



        else:
           self.steps += 1
           self.accuracy_e += self.accuracy_update
           self.accuracy_e = max(self.accuracy_e, 0)
           print(f'Step {self.steps}: Model not trained.')
           print()
           
        if self.accuracy_update!=0.0:
            self.accuracy_e += self.accuracy_update
            self.accuracy_e = max(self.accuracy_e, 0)
            print(f"Accuracy threshold updated to {self.accuracy_e:.4f}.")
            print()

        self.results['forces'] = F
        self.results['energy'] = E
        

    # Function to get features of an ASE atoms object
    def get_features(self, structures):
        features = np.array(self.descriptor.get_features(structures))
        print(np.shape(features))
        if (np.shape(features)[0])==1:
            features=features[0]
        else:
            features = features.sum(axis=1)
        return features





"""Modified ASE Langevin dynamics class with implementation of a temperature ramp."""

import numpy as np

from ase.md.md import MolecularDynamics
from ase.parallel import world, DummyMPI
from ase import units


class Langevin(MolecularDynamics):
    """Langevin (constant N, V, T) molecular dynamics."""

    # Helps Asap doing the right thing.  Increment when changing stuff:
    _lgv_version = 4

    def __init__(self, atoms, timestep, temperature=None, friction=None,
                 fixcm=True, *, temperature_K=None, trajectory=None,
                 logfile=None, loginterval=1, communicator=world,
                 rng=None, append_trajectory=False, tbegin=None, heatsteps=0):
        """
        Parameters:

        atoms: Atoms object
            The list of atoms.

        timestep: float
            The time step in ASE time units.

        temperature: float (deprecated)
            The desired temperature, in electron volt.

        temperature_K: float
            The desired temperature, in Kelvin.

        tbegin: float
            The starting temperature of the temperature ramp.

        heatsteps: float
            The number of MD steps of the temperature ramp.
            
        friction: float
            A friction coefficient, typically 1e-4 to 1e-2.

        fixcm: bool (optional)
            If True, the position and momentum of the center of mass is
            kept unperturbed.  Default: True.

        rng: RNG object (optional)
            Random number generator, by default numpy.random.  Must have a
            standard_normal method matching the signature of
            numpy.random.standard_normal.

        logfile: file object or str (optional)
            If *logfile* is a string, a file with that name will be opened.
            Use '-' for stdout.

        trajectory: Trajectory object or str (optional)
            Attach trajectory object.  If *trajectory* is a string a
            Trajectory will be constructed.  Use *None* (the default) for no
            trajectory.

        communicator: MPI communicator (optional)
            Communicator used to distribute random numbers to all tasks.
            Default: ase.parallel.world. Set to None to disable communication.

        append_trajectory: bool (optional)
            Defaults to False, which causes the trajectory file to be
            overwritten each time the dynamics is restarted from scratch.
            If True, the new structures are appended to the trajectory
            file instead.

        The temperature and friction are normally scalars, but in principle one
        quantity per atom could be specified by giving an array.

        RATTLE constraints can be used with these propagators, see:
        E. V.-Eijnden, and G. Ciccotti, Chem. Phys. Lett. 429, 310 (2006)

        The propagator is Equation 23 (Eq. 39 if RATTLE constraints are used)
        of the above reference.  That reference also contains another
        propagator in Eq. 21/34; but that propagator is not quasi-symplectic
        and gives a systematic offset in the temperature at large time steps.
        """
        if friction is None:
            raise TypeError("Missing 'friction' argument.")
        self.fr = friction
        self.temp = units.kB * self._process_temperature(temperature,
                                                         temperature_K, 'eV')
        self.fix_com = fixcm
        if communicator is None:
            communicator = DummyMPI()
        self.communicator = communicator
        if rng is None:
            self.rng = np.random
        else:
            self.rng = rng
        self.iterations=-1
        self.tbegin=tbegin
        self.tend=temperature_K
        self.tfric=1.0
        self.heatsteps=heatsteps
        MolecularDynamics.__init__(self, atoms, timestep, trajectory,
                                   logfile, loginterval,
                                   append_trajectory=append_trajectory)
        self.updatevars()

    def todict(self):
        d = MolecularDynamics.todict(self)
        d.update({'temperature_K': self.temp / units.kB,
                  'friction': self.fr,
                  'fixcm': self.fix_com})
        return d

    def set_temperature(self, temperature=None, temperature_K=None):
#        print('set temp')
        self.temp = units.kB * self._process_temperature(temperature,
                                                         temperature_K, 'eV')
        self.updatevars()

    def set_friction(self, friction):
        self.fr = friction
        self.updatevars()

    def set_timestep(self, timestep):
        self.dt = timestep
        self.updatevars()

    def updatevars(self):
#        print('update vars')
        dt = self.dt
        T = self.tfric*self.temp
#        print('Temp = ',T,self.temp)
        fr = self.fr
        masses = self.masses
        sigma = np.sqrt(2 * T * fr / masses)

        self.c1 = dt / 2. - dt * dt * fr / 8.
        self.c2 = dt * fr / 2 - dt * dt * fr * fr / 8.
        self.c3 = np.sqrt(dt) * sigma / 2. - dt**1.5 * fr * sigma / 8.
        self.c5 = dt**1.5 * sigma / (2 * np.sqrt(3))
        self.c4 = fr / 2. * self.c5

    def step(self, forces=None):
        atoms = self.atoms
        natoms = len(atoms)

        if forces is None:
            forces = atoms.get_forces(md=True)
#            print('\n Atoms.get_forces executed (forces was None).\n')

        self.iterations+=1
        if (self.tbegin!=None) and (self.iterations<=self.heatsteps):
           self.tfric=(self.tbegin+self.iterations*(self.tend-self.tbegin)/self.heatsteps)/self.tend
           self.updatevars()

        atoms.wrap()
        
        # This velocity as well as xi, eta and a few other variables are stored
        # as attributes, so Asap can do its magic when atoms migrate between
        # processors.
        self.v = atoms.get_velocities()

        self.xi = self.rng.standard_normal(size=(natoms, 3))
        self.eta = self.rng.standard_normal(size=(natoms, 3))

        # When holonomic constraints for rigid linear triatomic molecules are
        # present, ask the constraints to redistribute xi and eta within each
        # triple defined in the constraints. This is needed to achieve the
        # correct target temperature.
        for constraint in self.atoms.constraints:
            if hasattr(constraint, 'redistribute_forces_md'):
                constraint.redistribute_forces_md(atoms, self.xi, rand=True)
                constraint.redistribute_forces_md(atoms, self.eta, rand=True)

        self.communicator.broadcast(self.xi, 0)
        self.communicator.broadcast(self.eta, 0)

        # First halfstep in the velocity.
        self.v += (self.c1 * forces / self.masses - self.c2 * self.v +
                   self.c3 * self.xi - self.c4 * self.eta)

        # Full step in positions
        x = atoms.get_positions()
        if self.fix_com:
            old_com = atoms.get_center_of_mass()
        # Step: x^n -> x^(n+1) - this applies constraints if any.
        atoms.set_positions(x + self.dt * self.v + self.c5 * self.eta)
        if self.fix_com:
            atoms.set_center_of_mass(old_com)


        # recalc velocities after RATTLE constraints are applied
        self.v = (self.atoms.get_positions() - x -
                  self.c5 * self.eta) / self.dt
        forces = atoms.get_forces(md=True)
#        print('\n Atoms.get_forces executed (after RATTLE constraints...).\n')


        # Update the velocities
        self.v += (self.c1 * forces / self.masses - self.c2 * self.v +
                   self.c3 * self.xi - self.c4 * self.eta)

        if self.fix_com:  # subtract center of mass vel
            self.v -= self._get_com_velocity(self.v)

        # Second part of RATTLE taken care of here
        atoms.set_momenta(self.v * self.masses)

        return forces



"""Defines three default models implemented AGOX, that can be imported from falcon_md.models.agox_models"""

def GPR(atoms):
    """Standard Gaussian Progess Regression Model trained only on energies"""
    from agox.models.descriptors import Fingerprint
    from agox.models.GPR import GPR
    from agox.models.GPR.kernels import RBF, Noise
    from agox.models.GPR.kernels import Constant as C
    from agox.models.GPR.priors import Repulsive

    descriptor = Fingerprint.from_atoms(atoms)
    beta = 0.01
    k0 = C(beta, (beta, beta)) * RBF()
    k1 = C(1 - beta, (1 - beta, 1 - beta)) * RBF()
    kernel = C(5000, (1, 1e5)) * (k0 + k1) + Noise(0.01, (0.01, 0.01))

    model = GPR(
        descriptor=descriptor,
        kernel=kernel,
        prior=Repulsive(),
        use_ray=True
    )

    print("""\nYou are using the Gaussian Process Regression model implemented in AGOX. If you use this model in a publication, please cite:
Mads-Peter V. Christiansen, Nikolaj Rønne, Bjørk Hammer, "Atomistic Global Optimization X: A Python package for optimization of atomistic structures", J. Chem. Phys. 157, 054701 (2022),
<https://arxiv.org/abs/2204.01451>\n""")

    return model


def GPRForces(atoms):
    """Gaussian Progess Regression Model trained on forces and energies"""
    from agox.models.descriptors import Fingerprint
    from agox.models.GPR import SparseGPR
    from agox.models.GPR.kernels import RBF
    from agox.models.GPR.kernels import Constant as C

    descriptor = Fingerprint.from_atoms(atoms)
    kernel = C(1, (1, 100)) * RBF(20, (10, 30))

    model = SparseGPR(
        kernel=kernel,
        descriptor=descriptor,
        sparsifier=None,
        noise_E=0.01,  # eV/atom
        noise_F=0.05,  # eV/Å/atom
        force_data_filter="all",
        train_uncertainty=True,
        use_ray=True
    )
    
    print("""\nYou are using the Gaussian Process Regression model implemented in AGOX. If you use this model in a publication, please cite:
Mads-Peter V. Christiansen, Nikolaj Rønne, Bjørk Hammer, "Atomistic Global Optimization X: A Python package for optimization of atomistic structures", J. Chem. Phys. 157, 054701 (2022),
<https://arxiv.org/abs/2204.01451>\n""")

    return model


def SparseGPR(atoms):
    """Sparse Gaussian Progess Regression Model trained on forces and energies"""
    from agox.models.descriptors import Fingerprint
    from agox.models.GPR import SparseGPR
    from agox.models.GPR.kernels import RBF
    from agox.models.GPR.kernels import Constant as C
    from agox.utils.sparsifiers import CUR

    descriptor = Fingerprint.from_atoms(atoms)
    kernel = C(1, (1, 100)) * RBF(20, (10, 30))

    model = SparseGPR(
        kernel=kernel,
        descriptor=descriptor,
        sparsifier=CUR(1000),  # CUR-Sparsifizierung mit 1000 Punkten
        noise_E=0.01,  # eV/atom
        noise_F=0.05,  # eV/Å/atom
        force_data_filter="all",
        train_uncertainty=True,
        use_ray=True
    )
    
    print("""\nYou are using the SparseGPR model implemented in AGOX. If you use this model in a publication, please cite:
Mads-Peter V. Christiansen, Nikolaj Rønne, Bjørk Hammer, "Atomistic Global Optimization X: A Python package for optimization of atomistic structures", J. Chem. Phys. 157, 054701 (2022),
<https://arxiv.org/abs/2204.01451>\n""")

    return model



from ase.calculators.calculator import Calculator
from ase.calculators.singlepoint import SinglePointCalculator
from ase.io import write,Trajectory
from sklearn.cluster import KMeans
import numpy as np
import time
import copy

class FALCON_Static(Calculator):
    """
    Machine Learning Calculator using a static model.

    Parameters
    -----------
    model : falcon-md.models.model_base_class.ModelBaseClass or list of Models
        ML model to use or list of models to load.
 
    training_data : list of ASE atoms objects
        List of structures for training of the ML model.

    n_clusters : int
        Number of clusters/ML models used for training. Default: 1

    """

    implemented_properties = ['energy', 'forces']
    default_parameters = {}

    def __init__(self,
                 model,
                 training_data, 
                 n_clusters=1,
                 **kwargs):

        Calculator.__init__(self, **kwargs)

        self.steps = 0               # MD step count 
        self.initialize = True
        self.training_data = training_data
        self.n_clusters = n_clusters

        # Setup of the ML models
        if not type(model)==list:
            self.active_models = []
            self.descriptor = model.descriptor
            self.reloaded = False

        # Creation of number of  ML models that is defined with n_clusters.
            for i in range(n_clusters):
                model_copy = copy.deepcopy(model)
                setattr(self, f"model{i}", model_copy)
                self.active_models.append(getattr(self, f"model{i}"))

        else:
            self.models = model
            self.active_models = model
            self.descriptor = model[0].descriptor
            self.n_clusters = len(model)
            self.reloaded = True



    def calculate(self, atoms=None, properties=['energy', 'forces'], system_changes=['positions']):
        Calculator.calculate(self, atoms, properties, system_changes)

        print(f'\n============================ Step: {self.steps+1} ============================\n')
    
        print(f'Total structures in training data: {len(self.training_data)}')

        if self.initialize:
            start = time.time()

            print(f'{len(self.active_models)} active Models.')

            # Cluster all structures in training data using KMeans
            X = self.get_features(self.training_data)
            self.kmeans = KMeans(n_clusters=self.n_clusters, init="k-means++", n_init=10, random_state=np.random.randint(0, 10e6)).fit(X)
            self.labels = self.kmeans.labels_
            print(self.labels)
            self.clusters = {i: [] for i in range(self.n_clusters)}

            # Assign structures to their ML model.
            for structure, label in zip(self.training_data, self.labels):
                self.clusters[label].append(structure)

            for cluster_id, structures in self.clusters.items():
                print(f"Cluster {cluster_id} contains {len(structures)} objects.")
            end = time.time()
            t = end-start
            print(f"Clustering of {len(self.training_data)} took {t} s.")

            # Training of all active ML models.
            start = time.time()
            for i, model in enumerate(self.active_models):
                start_model = time.time()
                print(type(model))
                model.train(self.clusters[i])
                end_model = time.time()
                t_model = end_model - start_model
                print(f"Model {i} trained with {len(self.clusters[i])} structures ({t_model} s).")
                print()
            end = time.time()
            t = end-start
            print(f'Model trained with initial training data. ({t} s)')
            self.initialize = False

        # Prediction of energy, forces and uncertainties
        E = None
        F = None
        Fmax = None

        for i, model in enumerate(self.active_models):
            print(f'Model {i}...')
            start = time.time()
            E_model = model.predict_energy(atoms)
            end = time.time()
            t = end-start
            print(f'Predict energy ({t} s)')
            print(f'E = {E_model} (Model {i})')

            start = time.time()
            F_model = model.predict_forces(atoms)
            Fmax_model = np.max(np.linalg.norm(F_model, axis=1))
            end = time.time()
            t = end-start
            print(f'Predict forces ({t} s)')
            print(f'F = {Fmax_model} (Model {i})')
            print()

        if self.n_clusters>1:
            best_model = None
            Estd = float('inf')
            for i, model in enumerate(self.active_models):
                start = time.time()
                Estd_model = model.predict_uncertainty(atoms)
                end = time.time()
                t = end-start
                print(f'Predict energy_uncertainty ({t} s)')
                print(f'Estd = {Estd_model} (Model {i})')
                if Estd_model<Estd:
                    Estd = Estd_model
                    E = E_model
                    F = F_model
                    Fmax = Fmax_model
                    Fstdmax = Fstdmax_model
                    best_model = i
            print()
            print(f'Model {best_model} has the lowest uncertainty.')
            print(f'E = {E}')
            print(f'Estd = {Estd}')
            print(f'Fmax = {Fmax}')
            print(f'Fstd = {Fstdmax}')
            print()

        else:
            E = E_model
            F = F_model
        
        self.steps += 1
        print(f'Step {self.steps}: Model not trained.')
        print()
        self.results['forces'] = F
        self.results['energy'] = E
        

    # Function to get features of an ASE atoms object
    def get_features(self, structures):
        features = np.array(self.descriptor.get_features(structures)).sum(axis=1)
        return features





from typing import List
import numpy as np
from ase import Atoms


class ModelBaseClass():

    """Implementation of Model Base Class
    model needs to have a descriptor which is called in the FALCON main code as model.descriptor
    e.g. if one used the dscibe library: 
    self.descriptor=
    
    """


    def __init__(self):
        pass



    @property
    @abstractmethod
    def implemented_properties(self) -> list[str]:  # pragma: no cover
        """:obj: `list` of :obj: `str`: Implemented properties.
        Available properties are: 'energy', 'forces', 'uncertainty'

        Must be implemented in child class.
        """
        ...




    @abstractmethod
    def predict_energy(self, atoms: Atoms, **kwargs) -> float:  # pragma: no cover
        """Method for energy prediction.

        Note
        ----------
        Always include **kwargs when implementing this function.

        Parameters
        ----------
        atoms : ASE Atoms object
            The atoms object for which to predict the energy.

        Returns
        ----------
        E : float
            The energy prediction

        Must be implemented in child class.
        """

        # E = my_energy_prediction()

        return E
        



    @abstractmethod
    def train(self, training_data: List[Atoms], **kwargs) -> None:  # pragma: no cover
        """Method for model training.

        Note
        ----------
        Always include **kwargs when implementing this function.
        If your model is not trainable just write a method that does nothing

        Parameters
        ----------
        atoms : :obj: `list` of :obj: `ASE Atoms`
            List of ASE atoms objects to use as training data.
            All atoms must have a calculator with energy and other necessary properties set, such that
            it can be accessed by .get_* methods on the atoms.


        Must be implemented in child class.

        """

        # my_model_training()

        pass



    def predict_forces(self, atoms: Atoms, **kwargs) -> np.ndarray:
        """Method for forces prediction.

        Note
        ----------
        Always include **kwargs when implementing this function.

        Parameters
        ----------
        atoms : ASE Atoms object
            The atoms object for which to predict the forces.

        Returns
        ----------
        F : np.array
            The force prediction with shape (N,3), where N is len(atoms)

        """

        # F = my_forces_prediction()
        
        return F



    def predict_uncertainty(self, atoms: Atoms, **kwargs) -> float:
        """Method for energy uncertainty prediction.

        Parameters
        ----------
        atoms : ASE Atoms object
            The atoms object for which to predict the energy.

        Returns
        ----------
        E_unc : float
            The energy uncertainty prediction

        """

        # E_unc = my_energy_uncertainty_prediction()
        
        return E_unc



    def predict_uncertainty_forces(self, atoms: Atoms, **kwargs) -> np.ndarray:
        """Method for energy uncertainty prediction.

        Parameters
        ----------
        atoms : ASE Atoms object
            The atoms object for which to predict the energy.

        Returns
        ----------
        F_unc : np.array
            The force uncertainty prediction with shape (N,3) with N=len(atoms)

        """

        # F_unc = my_forces_uncertainty_prediction()

        return F_unc



class DescriptorBaseClass():

    """Implementation of Descriptor Base Class"""


    def __init__(self):
        pass

    def create_features(atoms: Atoms, **kwargs) -> np.ndarray:
        """
        Method to implement that does the calculation of the feature vector for a single structure.

        This method should not (generally) deal with being given a list of
        Atoms objects.

        Parameters
        ----------
        atoms : Atoms object

        Returns
        --------
        features : Features of shape [n_centers, n_features]
        """

        # features = my_feature_vector_calculation()

        return features


    def get_features(atoms: List[Atoms], **kwargs) -> np.ndarray:
        """
        Compute feature tensor for a list of Atoms objects.

        Parameters
        ----------
        atoms : list of ase.Atoms

        Returns
        -------
        features : np.ndarray
            Array of shape [n_structures, n_centers, n_features]
        """

        features = np.array([create_features(a, **kwargs) for a in atoms])

        return features




import os
import copy

def save_models(model_list, folder_path='models'):
    os.makedirs(folder_path, exist_ok=True)
    for idx, model in enumerate(model_list):
        model_path = os.path.join(folder_path, f"model_{idx}.h5")
        model.save(model_path)
        print(f"Saved model {idx} to {model_path}")


def load_models(basis_model, folder_path='models'):
    model_files = sorted([f for f in os.listdir(folder_path) if f.startswith("model_") and f.endswith(".h5")])
    models = []
    for file in model_files:
        full_path = os.path.join(folder_path, file)
        model = copy.deepcopy(basis_model)
        model.load(full_path)
        models.append(model)
        print(f"Loaded {full_path} into new model object")
    return models



import numpy as np
import matplotlib.pyplot as plt
from ase.io import read
import re

def plot_simulation(traj, outfile, timestep_fs=0.5, max_steps=None):
    """
    Plot potential energy over time with DFT-call markers from outfile.

    Parameters
    ----------
    traj : str
        Path to ASE trajectory file.

    outfile : str
        Path to the FALCON outputfile containing training events.

    timestep_fs : float
        Time per MD step in fs.

    max_steps : int, optional
        Limit number of trajectory steps to read. Default: None (read all).
    """

    # Read trajectory
    traj = read(traj, index=f':{max_steps}' if max_steps else ':')
    energies = [atoms.get_potential_energy() for atoms in traj]
    times_ps = np.arange(len(energies)) * (timestep_fs / 1000)  # Convert fs to ps

    # Parse outputfile for DFT call steps (MD steps where training occurred)
    pattern = r"Step\s+(\d+):\s+Model\s+\d+\s+trained\s+with\s+\d+\s+Structures\s+\([\d.]+ s\).\s+\((\d+)\s"
    with open(outfile, 'r') as f:
        text = f.read()

    matches = re.findall(pattern, text)
    dft_md_steps = [int(md_step) for md_step, _ in matches if int(md_step) < len(energies)]
    dft_md_steps = np.array(dft_md_steps)

    # Plotting
    plt.figure(figsize=(10, 5))
    plt.plot(times_ps, energies, linewidth=2, alpha=0.6, color='#E41A1C', label="Energy")
    plt.scatter(times_ps[dft_md_steps], np.array(energies)[dft_md_steps],
                color='k', s=6, label='DFT call', zorder=5)

    plt.xlabel('Time (ps)')
    plt.ylabel('Energy (eV)')
    plt.legend(handlelength=0.1, markerscale=2, frameon=False)
    plt.tight_layout()
    plt.show()



import numpy as np
import matplotlib.pyplot as plt
from ase.geometry.analysis import Analysis
from ase.io import read
from ase.build import make_supercell


def plot_rdf(filename: str,
             index_solid: str,
             index_liquid: str,
             rmax: float = 8.0,
             nbins: int = 200):
    """
    Compute and plot average RDFs for solid and liquid segments of a trajectory.

    Parameters
    ----------
    filename : str
        Path to trajectory file.
    index_solid : str
        ASE slice string for solid frames (e.g. '4000:5000:200').
    index_liquid : str
        ASE slice string for liquid frames (e.g. '49000:50000:200').
    rmax : float, optional
        Maximum distance in Å for RDF calculation (default: 8.0).
    nbins : int, optional
        Number of RDF bins (default: 200).
    """

    def compute_avg_rdf(traj):
        rdf_accum = []
        for atoms in traj:
            sc = make_supercell(atoms, np.identity(3) * 3)
            rdf = Analysis(sc).get_rdf(rmax=rmax, nbins=nbins)[0]
            rdf_accum.append(rdf)
        return np.mean(rdf_accum, axis=0)

    traj_solid = read(filename, index=index_solid)
    traj_liquid = read(filename, index=index_liquid)

    r = np.linspace(0.0, rmax, nbins)
    rdf_solid = compute_avg_rdf(traj_solid)
    rdf_liquid = compute_avg_rdf(traj_liquid)

    colors = ["#CCBB44", "#EE6677"]
    plt.figure(figsize=(5, 5))
    plt.plot(r, rdf_solid, color=colors[0], linewidth=2, label='Solid')
    plt.plot(r, rdf_liquid, color=colors[1], linewidth=2, label='Liquid')

    plt.ylabel('g(r)')
    plt.xlabel(r'r ($\mathrm{\AA}$)')
    plt.yticks([])
    xticks_major = np.arange(1, rmax, 1)
    xticks_minor = np.arange(1.5, rmax, 1)

    plt.xticks(xticks_major)
    plt.xlim(1, rmax)

    ax = plt.gca()
    ax.set_xticks(xticks_minor, minor=True)

    plt.legend(frameon=True, fontsize=12, markerscale=2/3)
    plt.tight_layout()
    plt.show()



from setuptools import setup
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text(encoding="utf-8")

setup(
    name='falcon_md',
    version='1.3.0',
    description='The python distribution for the FALCON on-the-fly Machine Learning ab initio Molecular Dynamics code',
    long_description = long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/thequantumchemist/falcon',
    author='Noah Felis and  Wilke Dononelli',
    author_email='wido@uni-bremen.de',
    license='GPL-3.0',
    packages=['falcon_md','falcon_md/utils', 'falcon_md/models', 'falcon_md/structures', 'falcon_md/utils/analysis'],
    package_data={
        'falcon_md/structures': ['*.xyz', '*.traj'],
    },
    include_package_data=True,
    python_requires=">=3.8",
    install_requires=['agox',
                      'numpy',
                      'ase',
                      'pytest'
                      ],

    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',

        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.8',
    ],
)



import os
import copy
from ase.io import read, write, Trajectory
from sklearn.cluster import KMeans
import numpy as np

def get_uncertain_points(model, trajectory, n_clusters=50, points_per_cluster=1, threshold=1e9, write_structures=True, path='Uncertain_Strcutures.traj'):
    traj = read(trajectory, index=':')


    uncertainties=[]

    for i, atoms in enumerate(traj):
        best_unc = float('inf')
        for i, m in enumerate(model):
            unc = m.predict_uncertainty(atoms)
            if unc < best_unc:
                best_unc = unc
        uncertainties.append(best_unc)

    descriptor = model[0].descriptor
    X = np.array(descriptor.get_features(traj)).sum(axis=1)
    kmeans = KMeans(n_clusters=n_clusters, init="k-means++", n_init=10, random_state=np.random.randint(0, 10e6)).fit(X)
    labels = kmeans.labels_
    clusters = [[] for _ in range(n_clusters)]
    cluster_unc = [[] for _ in range(n_clusters)]
    idx = [[] for _ in range(n_clusters)]

    for i, (structure, label, uncertainty) in enumerate(zip(traj, labels, uncertainties)):
        clusters[label].append(structure)
        cluster_unc[label].append(uncertainty)
        idx[label].append(i)

    uncertain_structures = []

    print()
    print(f"Most uncertain points in the analyzed Trajectory:")
    for cluster_idx in range(len(clusters)):
        unc_values = cluster_unc[cluster_idx]
        sorted_indices = sorted(range(len(unc_values)), key=lambda i: unc_values[i], reverse=True)
        top_n_indices = sorted_indices[:points_per_cluster]


        for i in top_n_indices:
            if cluster_unc[cluster_idx][i]>threshold:
                print(f"MD step: {idx[cluster_idx][i]},  Uncertainty: {cluster_unc[cluster_idx][i]} eV, Cluster: {cluster_idx}")
                uncertain_structures.append(clusters[cluster_idx][i])

    if write_structures:
        with Trajectory(path, mode='w') as traj:
                            for structure in uncertain_structures:
                                traj.write(structure)



from ase.io import read
from pathlib import Path

_structures_dir = Path(__file__).parent

def load_structure(prefix):
    """Load .xyz or .traj file of tutorial structures by prefix from 'falcon_md/structures' directory (without extension)."""
    
    for ext in [".xyz", ".traj"]:
        path = _structures_dir / f"{prefix}{ext}"
        if path.exists():
            print(f"Structure {path} loaded.")
            return read(path)
    raise FileNotFoundError(f"No structure found for '{prefix}' in {_structures_dir}")
